DUO_UNIVERSAL_PROMPT_FACTOR_DUO_PUSH = "Duo Push"
DUO_UNIVERSAL_PROMPT_FACTOR_PASSCODE = "Passcode"
DUO_UNIVERSAL_PROMPT_FACTOR_PHONE_CALL = "Phone Call"
DUO_UNIVERSAL_PROMPT_FACTOR_WEBAUTHN = "WebAuthn Security Key"
